#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	vec4 fragPosLightSpace;
	mat3 TBN;
} vert;

out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;

uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;
uniform sampler2D normalMap;
uniform sampler2D shadowMap;
uniform samplerCube cubemapSampler;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL[10];

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL[10];

uniform int dirLightNumber;
uniform int pointLightNumber;

uniform float time;
uniform bool normalsActive;


vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

float calcShadow(vec4 fragPosLightSpace, int i){

	float bias = max(0.05 * (1.0 - dot(vert.normal_world, dirL[i].direction)), 0.005);

	vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
   	projCoords = projCoords * 0.5 + 0.5;
    	float closestDepth = texture(shadowMap, projCoords.xy).r; 
    	float currentDepth = projCoords.z;

 	float shadow = 0.0;
    	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
    	for(int x = -1; x <= 1; ++x)
    	{
        	for(int y = -1; y <= 1; ++y)
        	{
            		float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
            		shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;        
        	}    
    	}
    	shadow /= 9.0;
    
    	if(projCoords.z > 1.0)
        	shadow = 0.0;
        
    	return shadow;
}

float cosineInterpolate(float a, float b, float x){
	float ft = x * 3.1415927;
	float f = (1 - cos(ft)) * 0.5;

	return  a*(1-f) + b*f;
}

float rand(float x, float y, float z){
    return fract(sin(dot(vec3(x,y,z),
                         vec3(12.9898,78.233,64.657)))*
        43758.5453123);
}

float smootheNoise(float x, float y,float z){

	float result;
    result += ( rand(x-1, y-1,z+1)+rand(x+1, y-1,z+1)+rand(x-1, y+1,z+1)+rand(x+1, y+1,z+1)+rand(x-1, y-1,z-1)+rand(x+1, y-1,z-1)+rand(x-1, y+1,z-1)+rand(x+1, y+1,z-1) ) / 32;
    result += ( rand(x-1, y, z)  +rand(x+1, y, z)  + rand(x, y-1, z)  + rand(x, y+1, z) + rand(x, y, z-1)  + rand(x, y, z+1)) /  16;
	result += ( rand(x+1, y, z+1)  +rand(x+1, y, z-1)  + rand(x+1, y+1, z)  + rand(x+1, y-1, z)) /  8;
	result += ( rand(x, y-1, z-1)  +rand(x, y-1, z+1)  + rand(x, y+1, z-1)  + rand(x, y+1, z+1)) /  8;
	result += ( rand(x-1, y, z+1)  +rand(x-1, y, z-1)  + rand(x-1, y+1, z)  + rand(x-1, y-1, z)) /  8;
    result +=  rand(x, y,z) / 4;
    return result;

}

float interpolatedNoise(float x, float y, float z){
	int intX = int(x);
	float fracX = x-intX;

	int intY = int(y);
	float fracY = y-intY;

	int intZ = int(z);
	float fracZ = z-intZ;
	
	float v1 = smootheNoise(intX,intY,intZ);
    float v2 = smootheNoise(intX + 1, intY,intZ);
    float v3 = smootheNoise(intX,     intY + 1,intZ);
    float v4 = smootheNoise(intX + 1, intY + 1,intZ);

	float v5 = smootheNoise(intX,intY,intZ+1);
    float v6 = smootheNoise(intX + 1, intY,intZ+1);
    float v7 = smootheNoise(intX,     intY + 1,intZ+1);
    float v8 = smootheNoise(intX + 1, intY + 1,intZ+1);


    float i1 = cosineInterpolate(v1 , v2 , fracX);
    float i2 = cosineInterpolate(v3 , v4 , fracX);

    float i3 = cosineInterpolate(v5 , v6 , fracX);
    float i4 = cosineInterpolate(v7 , v8 , fracX);
	
    float i5 = cosineInterpolate(i1 , i3 , fracZ);
    float i6 = cosineInterpolate(i2 , i4 , fracZ);
	
    return cosineInterpolate(i5 , i6 , fracY);

}

float perlinNoise(float x, float y, float z){
	
	float t = 0;
	for(int i = 0; i < 1; i++){
		float freq = pow(2,i);
		float amp = pow(0.4,i);
		t += (interpolatedNoise(x * freq, y * freq, z * freq) * amp);
	}
	return t/2;
}

vec3 fire(float x, float y, float t){
	float intensity1 = perlinNoise(x*100, y*100, t);


	vec3 col = vec3(1,abs(intensity1-1),abs(intensity1-1)/2);	
	
	return col;
}
vec3 calcProcedularlNormal(float x, float y, float t){
	float nx = distance(fire(x,y,t),fire(x+1,y,t))-0.2;
	float ny = distance(fire(x,y,t),fire(x,y+1,t))-0.2;
	vec2 a = normalize(vec2(nx,ny));
	float nz = 5;
	
	return normalize(vec3(a.x,a.y,nz));
}

vec3 calcProceduralSpecCol(float x, float y, float t){
	return vec3(fire(x,y,t).x/5);
}

void main() {	
	vec3 n;
	
	if(normalsActive){
		n = calcProcedularlNormal(vert.uv.x,vert.uv.y,1);
		n = n * 2.0 - 1.0;
		n = normalize(vert.TBN * n);
	}else{
		n = normalize(vert.normal_world);
	}

	vec3 v = normalize(camera_world - vert.position_world);
	
	//vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	vec3 texColor = fire(vert.uv.x,vert.uv.y,time);

	//vec3 specColor = texture(specularTexture, vert.uv).rgb;
	vec3 specColor = calcProceduralSpecCol(vert.uv.x,vert.uv.y,time);
	
	vec3 cubeColor = texture(cubemapSampler,reflect(normalize(vert.position_world-camera_world),normalize(n))).rgb;

	color = vec4(((cubeColor*specColor + texColor) * materialCoefficients.x)/2, 1); // ambient with cube map
	//color = vec4((specColor + texColor * materialCoefficients.x)/2, 1); // ambient with cube map

	
	// add directional light contribution
	for(int i = 0; i < dirLightNumber; i++){
		//color.rgb = vec3(0,1,0);

		float shadow = calcShadow(vert.fragPosLightSpace, i);
		color.rgb += (1.0 - shadow) * phong(n, -dirL[i].direction, v, dirL[i].color * texColor, materialCoefficients.y, dirL[i].color  * specColor, materialCoefficients.z, specularAlpha, false, vec3(0));
	}	
		

	// add point light contribution
	//for(int i = 0; i < pointLightNumber; i++){
	//	color.rgb += phong(n, pointL[i].position - vert.position_world, v, pointL[i].color * texColor, materialCoefficients.y, pointL[i].color * specColor, materialCoefficients.z, specularAlpha, true, pointL[i].attenuation);
	//}

}



